import path from 'path'
import fs from 'fs'
import fse from 'fs-extra'

import { rollup } from 'rollup'
import json from '@rollup/plugin-json'
import { nodeResolve } from '@rollup/plugin-node-resolve'
import replace from '@rollup/plugin-replace'
import Vue from 'unplugin-vue/rollup'
import commonjs from '@rollup/plugin-commonjs';
import esbuild from 'rollup-plugin-esbuild'
import styles from "rollup-plugin-styles";
import polyfillNode from 'rollup-plugin-polyfill-node';
import uglify from 'uglify-js'
import buildConf from './config.js'
import { 
  scriptPaths, 
  useUiPackageJson,
  logError, 
  writeFile, 
} from './utils.js'

// import { add } from 'lodash-es'
// import postcss from 'rollup-plugin-postcss'

const { version } = useUiPackageJson()
const { __dirname } = scriptPaths(import.meta.url)

const rollupPlugins = [
  replace({
    preventAssignment: false,
    values: {
      __UI_VERSION__: `'${ version }'`
    }
  }),
  nodeResolve({
    extensions: ['.js', '.ts'],
    preferBuiltins: false,
    browser: true,
  }),

  commonjs({
    // avoids error "default" is not exported by ".../extend/index.js", imported by ".../unified/lib/index.js"
    include: /\/node_modules\//,
    namedExports: {
      'extend': ['default'],
    },
  }),

  Vue({ 
    // css: false, 
    style: {
      preprocessLang: 'scss',
      preprocessOptions: {
        additionalData: `
          @import '${__dirname }/../node_modules/quasar/src/css/variables.sass';
        `,
      }
    },
  }),

  esbuild({
    // bundle: true,
    // platform: 'node',
    platform: 'browser',
    target: 'esnext',
    sourcemap: true,
    // minify: true,
  }),

  styles(),

  polyfillNode(), // avoids warning "crypto" is imported by "...nanoid/index.js", but could not be resolved – treating it as an external dependency.

  // postcss({
  //   extensions: ['.css', '.scss'],
  //   extract: false, // Set to true to extract CSS into a separate file
  //   // minimize: true,
  //   use: ['sass'],
  // }),

  json(),
]

const builds = [
  {
    rollup: {
      input: {
        input: pathResolve('../src/index.esm.js')
      },
      output: {
        dir: pathResolve('../dist/esm'),
        format: 'es',
        preserveModules: true, // Preserve module structure
      }
    },
    build: {
      unminified: true,
      // minified: true
    }
  },
  /** /
  {
    rollup: {
      input: {
        input: pathResolve('../src/index.esm.js')
      },
      output: {
        file: pathResolve('../dist/index.esm.js'),
        format: 'es',
        inlineDynamicImports: true, // allow dynamic imports()
      }
    },
    build: {
      unminified: true,
      // minified: true
    }
  },
  {
    rollup: {
      input: {
        input: pathResolve('../src/index.common.js')
      },
      output: {
        file: pathResolve('../dist/index.common.js'),
        format: 'cjs',
        inlineDynamicImports: true, // allow dynamic imports()
      }
    },
    build: {
      unminified: true,
      // minified: true
    }
  },
  /**/
  // {
  //   rollup: {
  //     input: {
  //       input: pathResolve('../src/index.umd.js')
  //     },
  //     output: {
  //       name: 'QMilkdown',
  //       file: pathResolve('../dist/index.umd.js'),
  //       format: 'umd',
  //       inlineDynamicImports: true, // allow dynamic imports()
  //     }
  //   },
  //   build: {
  //     unminified: true,
  //     minified: true,
  //     minExt: true
  //   }
  // }
]

// Add your asset folders here, if needed
// addAssets(builds, 'icon-set', 'iconSet')
// addAssets(builds, 'lang', 'lang')

build(builds)

/**
 * Helpers
 */

function pathResolve (_path) {
  return path.resolve(__dirname, _path)
}

// eslint-disable-next-line no-unused-vars
function addAssets (builds, type, injectName) {
  const
    files = fs.readdirSync(pathResolve('../../ui/src/components/' + type)),
    plugins = [],
    outputDir = pathResolve(`../dist/${type}`)

    fse.mkdirp(outputDir)

  files
    .filter(file => file.endsWith('.js'))
    .forEach(file => {
      const name = file.substr(0, file.length - 3).replace(/-([a-z])/g, g => g[1].toUpperCase())
      builds.push({
        rollup: {
          input: {
            input: pathResolve(`../src/components/${type}/${file}`),
            plugins
          },
          output: {
            file: addExtension(pathResolve(`../dist/${type}/${file}`), 'umd'),
            format: 'umd',
            name: `QMilkdown.${injectName}.${name}`
          }
        },
        build: {
          minified: true
        }
      })
    })
}

function build (builds) {
  return Promise
    .all(builds.map(genConfig).map(buildEntry))
    // .catch(logError)
    .catch(err => {
      console.error(err)
      process.exit(1)
    })
}

function genConfig (opts) {
  Object.assign(opts.rollup.input, {
    plugins: rollupPlugins,
    external: [ 
      'vue', 
      'quasar',
      'node:crypto', // "node:crypto" is imported by "../../../node_modules/.pnpm/nanoid@5.0.9/node_modules/nanoid/index.js", but could not be resolved – treating it as an external dependency.
    ],
    // logLevel: 'debug',
    onwarn(warning, warn) {
      // Suppress circular dependency warnings for specific packages
      // console.log('!!!', warning)
      // if (warning.code === 'CIRCULAR_DEPENDENCY' && warning.importer?.includes(
      //   'atomico'
      // )) {
      //   return;
      // }
      warn(warning);
    },
  })

  Object.assign(opts.rollup.output, {
    banner: buildConf.banner,
    globals: { vue: 'Vue', quasar: 'Quasar' }
  })

  return opts
  // .console('rollup opts', JSON.stringify)
}

function addExtension (filename, ext = 'min') {
  const insertionPoint = filename.lastIndexOf('.')
  return `${filename.slice(0, insertionPoint)}.${ext}${filename.slice(insertionPoint)}`
}

async function buildEntry (config) {
  const bundle = await rollup(config.rollup.input)

  if (config.rollup.output.dir) {
    bundle.write(config.rollup.output)
    return
  }

  const { output } = bundle.generate(config.rollup.output)

  let code = config.rollup.output.format === 'umd'
  ? injectVueRequirement(output[0].code)
  : output[0].code

  code = config.build.unminified
    ? writeFile(config.rollup.output.file, code)
    : code

  if (!config.build.minified) {
    return code
  }

  const minified = uglify.minify(code, {
    compress: {
      pure_funcs: ['makeMap']
    }
  })

  if (minified.error) {
    throw minified.error
  }

  await writeFile(
    config.build.minExt === true
      ? addExtension(config.rollup.output.file)
      : config.rollup.output.file,
    buildConf.banner + minified.code,
    true
  )
}

function injectVueRequirement (code) {
  // eslint-disable-next-line
  const index = code.indexOf(`Vue = Vue && Vue.hasOwnProperty('default') ? Vue['default'] : Vue`)

  if (index === -1) {
    return code
  }

  const checkMe = ` if (Vue === void 0) {
    console.error('[ Quasar ] Vue is required to run. Please add a script tag for it before loading Quasar.')
    return
  }
  `

  return code.substring(0, index - 1) +
    checkMe +
    code.substring(index)
}
